<?php

/**
 * UpdatePiggybank.php
 * Copyright (c) 2020 james@firefly-iii.org
 *
 * This file is part of Firefly III (https://github.com/firefly-iii).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

declare(strict_types=1);

namespace FireflyIII\TransactionRules\Actions;

use FireflyIII\Models\PiggyBank;
use FireflyIII\Models\RuleAction;
use FireflyIII\Models\Transaction;
use FireflyIII\Models\TransactionJournal;
use FireflyIII\Models\TransactionType;
use FireflyIII\Repositories\PiggyBank\PiggyBankRepositoryInterface;
use FireflyIII\User;
use Log;

/**
 * Class UpdatePiggybank
 */
class UpdatePiggybank implements ActionInterface
{
    private RuleAction $action;

    /**
     * TriggerInterface constructor.
     *
     * @param  RuleAction  $action
     */
    public function __construct(RuleAction $action)
    {
        $this->action = $action;
    }

    /**
     * @inheritDoc
     */
    public function actOnArray(array $journal): bool
    {
        Log::debug(sprintf('Triggered rule action UpdatePiggybank on journal #%d', $journal['transaction_journal_id']));

        // refresh the transaction type.
        $user = User::find($journal['user_id']);
        /** @var TransactionJournal $journalObj */
        $journalObj                       = $user->transactionJournals()->find($journal['transaction_journal_id']);
        $type                             = TransactionType::find((int)$journalObj->transaction_type_id);
        $journal['transaction_type_type'] = $type->type;

        if (TransactionType::TRANSFER !== $journal['transaction_type_type']) {
            Log::info(sprintf('Journal #%d is a "%s" so skip this action.', $journal['transaction_journal_id'], $journal['transaction_type_type']));

            return false;
        }

        $piggyBank = $this->findPiggyBank($user);
        if (null === $piggyBank) {
            Log::info(
                sprintf('No piggy bank named "%s", cant execute action #%d of rule #%d', $this->action->action_value, $this->action->id, $this->action->rule_id)
            );

            return false;
        }

        Log::debug(sprintf('Found piggy bank #%d ("%s")', $piggyBank->id, $piggyBank->name));

        /** @var Transaction $source */
        /** @var Transaction $destination */
        $source      = $journalObj->transactions()->where('amount', '<', 0)->first();
        $destination = $journalObj->transactions()->where('amount', '>', 0)->first();

        if ((int)$source->account_id === (int)$piggyBank->account_id) {
            Log::debug('Piggy bank account is linked to source, so remove amount from piggy bank.');
            $this->removeAmount($piggyBank, $journalObj, $destination->amount);

            return true;
        }
        if ((int)$destination->account_id === (int)$piggyBank->account_id) {
            Log::debug('Piggy bank account is linked to source, so add amount to piggy bank.');
            $this->addAmount($piggyBank, $journalObj, $destination->amount);

            return true;
        }
        Log::info(
            sprintf(
                'Piggy bank is not linked to source ("#%d") or destination ("#%d"), so no action will be taken.',
                $source->account_id,
                $destination->account_id
            )
        );

        return false;
    }

    /**
     * @param  User  $user
     *
     * @return PiggyBank|null
     */
    private function findPiggyBank(User $user): ?PiggyBank
    {
        return $user->piggyBanks()->where('piggy_banks.name', $this->action->action_value)->first();
    }

    /**
     * @param  PiggyBank  $piggyBank
     * @param  TransactionJournal  $journal
     * @param  string  $amount
     * @return void
     */
    private function removeAmount(PiggyBank $piggyBank, TransactionJournal $journal, string $amount): void
    {
        $repository = app(PiggyBankRepositoryInterface::class);
        $repository->setUser($journal->user);

        // how much can we remove from this piggy bank?
        $toRemove = $repository->getCurrentAmount($piggyBank);
        Log::debug(sprintf('Amount is %s, max to remove is %s', $amount, $toRemove));

        // if $amount is bigger than $toRemove, shrink it.
        $amount = -1 === bccomp($amount, $toRemove) ? $amount : $toRemove;
        Log::debug(sprintf('Amount is now %s', $amount));

        // if amount is zero, stop.
        if (0 === bccomp('0', $amount)) {
            app('log')->warning('Amount left is zero, stop.');

            return;
        }

        // make sure we can remove amount:
        if (false === $repository->canRemoveAmount($piggyBank, $amount)) {
            app('log')->warning(sprintf('Cannot remove %s from piggy bank.', $amount));

            return;
        }
        Log::debug(sprintf('Will now remove %s from piggy bank.', $amount));

        $repository->removeAmount($piggyBank, $amount, $journal);
    }

    /**
     * @param  PiggyBank  $piggyBank
     * @param  TransactionJournal  $journal
     * @param  string  $amount
     * @return void
     */
    private function addAmount(PiggyBank $piggyBank, TransactionJournal $journal, string $amount): void
    {
        $repository = app(PiggyBankRepositoryInterface::class);
        $repository->setUser($journal->user);

        // how much can we add to the piggy bank?
        $toAdd = bcsub($piggyBank->targetamount, $repository->getCurrentAmount($piggyBank));
        Log::debug(sprintf('Max amount to add to piggy bank is %s, amount is %s', $toAdd, $amount));

        // update amount to fit:
        $amount = -1 === bccomp($amount, $toAdd) ? $amount : $toAdd;
        Log::debug(sprintf('Amount is now %s', $amount));

        // if amount is zero, stop.
        if (0 === bccomp('0', $amount)) {
            app('log')->warning('Amount left is zero, stop.');

            return;
        }

        // make sure we can add amount:
        if (false === $repository->canAddAmount($piggyBank, $amount)) {
            app('log')->warning(sprintf('Cannot add %s to piggy bank.', $amount));

            return;
        }
        Log::debug(sprintf('Will now add %s to piggy bank.', $amount));

        $repository->addAmount($piggyBank, $amount, $journal);
    }
}
